@@ -350,68 +350,88 @@ xfce_shortcuts_grabber_ungrab_all (XfceShortcutsGrabber *grabber)
 
 
 static gboolean
-get_entries_for_keyval (GdkKeymap     *keymap,
-                        gint           group,
-                        guint          keyval,
-                        GdkKeymapKey **keys,
-                        guint         *n_keys)
+get_entries_for_keyval (GdkKeymap        *keymap,
+                        gint              group,
+                        guint             keyval,
+                        GdkModifierType   modifiers,
+                        GdkKeymapKey    **keys_out,
+                        guint            *n_keys_out)
 {
-  GdkKeymapKey *keys1;
-  gint n_keys1;
+  GdkKeymapKey *keys;
+  gint n_keys;
 
-  *keys = NULL;
-  *n_keys = 0;
+  *keys_out = NULL;
+  *n_keys_out = 0;
 
    /* Get all keys generating keyval */
-  if (!gdk_keymap_get_entries_for_keyval (keymap, keyval, &keys1, &n_keys1))
+  if (!gdk_keymap_get_entries_for_keyval (keymap, keyval, &keys, &n_keys))
     {
       TRACE ("Got no keys for keyval");
       return FALSE;
     }
 
-  if (G_UNLIKELY (n_keys1 <= 0))
+  if (G_UNLIKELY (n_keys <= 0))
     {
-      g_free (keys1);
+      g_free (keys);
       return FALSE;
     }
 
   /* Filter keys by group */
   {
-    gboolean group0_only;
-    gint i, n_matches;
+    gboolean group0_only = TRUE;
 
     /* For keys such as F12:
      *   keys1[i].group is always 0 (even if n_keys1 >= 2)
      *   and thus n_matches will be zero if group != 0 */
-
-    group0_only = TRUE;
-    n_matches = 0;
-    for (i = 0; i < n_keys1; i++)
+    for (gint i = 0; i < n_keys; i++)
       {
-        group0_only &= (keys1[i].group == 0) ? TRUE : FALSE;
-        if (keys1[i].group == group)
-          n_matches++;
+        if (keys[i].group != 0)
+          {
+            group0_only = FALSE;
+            break;
+          }
       }
 
-    if (!group0_only || n_matches != 0)
+    if (!group0_only)
       {
         /* Remove keys that do not match the group*/
-        for (i = 0; i < n_keys1;)
-          if (keys1[i].group == group)
+        for (gint i = 0; i < n_keys;)
+          if (keys[i].group == group)
             i++;
           else
-            keys1[i] = keys1[--n_keys1];
+            keys[i] = keys[--n_keys];
       }
   }
 
-  if (G_UNLIKELY (n_keys1 == 0))
+  /* Filter keys by level */
+  {
+    guint keyval_trans;
+    GdkModifierType modifiers_levelled;
+
+    /* Remove keys that do not match the level for the given modifiers */
+    for (gint i = 0; i < n_keys;)
+      {
+        modifiers_levelled = modifiers;
+        if (keys[i].level == 0)
+          modifiers_levelled &= ~(GDK_SHIFT_MASK | GDK_MOD5_MASK);
+
+        gdk_keymap_translate_keyboard_state (keymap, keys[i].keycode, modifiers_levelled,
+                                             keys[i].group, &keyval_trans, NULL, NULL, NULL);
+        if (keyval == keyval_trans)
+          i++;
+        else
+          keys[i] = keys[--n_keys];
+      }
+  }
+
+  if (G_UNLIKELY (n_keys == 0))
     {
-      g_free (keys1);
+      g_free (keys);
       return FALSE;
     }
 
-  *keys = keys1;
-  *n_keys = n_keys1;
+  *keys_out = keys;
+  *n_keys_out = n_keys;
   return TRUE;
 }
 
@@ -483,7 +503,7 @@ xfce_shortcuts_grabber_regrab_all (XfceShortcutsGrabber *grabber)
 
     if (!map_virtual_modifiers (keymap, key->modifiers, &non_virtual_modifiers))
       continue;
-    if (!get_entries_for_keyval (keymap, group, key->keyval, &keys, &n_keys))
+    if (!get_entries_for_keyval (keymap, group, key->keyval, key->modifiers, &keys, &n_keys))
       continue;
 
     already_grabbed = TRUE;
@@ -624,7 +644,7 @@ xfce_shortcuts_grabber_grab (XfceShortcutsGrabber *grabber, XfceKey *key)
 
   if (!map_virtual_modifiers (keymap, key->modifiers, &non_virtual_modifiers))
     return;
-  if (!get_entries_for_keyval (keymap, group, key->keyval, &keys, &n_keys))
+  if (!get_entries_for_keyval (keymap, group, key->keyval, key->modifiers, &keys, &n_keys))
     return;
 
 #ifdef DEBUG_TRACE
@@ -790,6 +810,28 @@ find_event_key (const gchar                *shortcut,
 
 
 
+static gboolean
+is_modifier_key (struct EventKeyFindContext context)
+{
+  if (context.modifiers == 0)
+    {
+      switch (context.keyval)
+        {
+          case GDK_KEY_Control_L:
+          case GDK_KEY_Control_R:
+          case GDK_KEY_Alt_L:
+          case GDK_KEY_Alt_R:
+          case GDK_KEY_Super_L:
+          case GDK_KEY_Super_R:
+            return TRUE;
+        }
+    }
+
+  return FALSE;
+}
+
+
+
 static GdkFilterReturn
 xfce_shortcuts_grabber_event_filter (GdkXEvent *gdk_xevent,
                                      GdkEvent  *event,
@@ -805,6 +847,11 @@ xfce_shortcuts_grabber_event_filter (GdkXEvent *gdk_xevent,
   gchar                      *raw_shortcut_name;
   gint                        timestamp;
 
+  /* We only activate single modifier keys on release event to allow combinations
+   * such as Super to open a menu and Super+T to open a terminal: see
+   * https://gitlab.xfce.org/xfce/libxfce4ui/-/issues/1 */
+  static gboolean             single_modifier_down = FALSE;
+
   g_return_val_if_fail (XFCE_IS_SHORTCUTS_GRABBER (grabber), GDK_FILTER_CONTINUE);
 
   xevent = (XEvent *) gdk_xevent;
@@ -825,7 +872,7 @@ xfce_shortcuts_grabber_event_filter (GdkXEvent *gdk_xevent,
     }
 #endif
 
-  if (xevent->type != KeyPress)
+  if (xevent->type != KeyPress && !(xevent->type == KeyRelease && single_modifier_down))
     return GDK_FILTER_CONTINUE;
 
   context.result = NULL;
@@ -838,6 +885,10 @@ xfce_shortcuts_grabber_event_filter (GdkXEvent *gdk_xevent,
   mod_mask = gtk_accelerator_get_default_mod_mask ();
   modifiers = xevent->xkey.state;
 
+  /* Remove modifier added to single modifier key on release event */
+  if (xevent->type == KeyRelease && single_modifier_down)
+    modifiers = 0;
+
   gdk_keymap_translate_keyboard_state (keymap, xevent->xkey.keycode,
                                        modifiers,
                                        grabber->priv->xkbStateGroup,
@@ -890,10 +941,21 @@ xfce_shortcuts_grabber_event_filter (GdkXEvent *gdk_xevent,
                      (GHRFunc) (void (*)(void)) find_event_key,
                      &context);
 
+  single_modifier_down = FALSE;
   if (G_LIKELY (context.result != NULL))
-    /* We had a positive match */
-    g_signal_emit_by_name (grabber, "shortcut-activated",
-                           context.result, timestamp);
+    {
+      /* We had a positive match */
+      if (xevent->type == KeyPress && is_modifier_key (context))
+        {
+          /* Will be activated on release event */
+          single_modifier_down = TRUE;
+        }
+      else
+        {
+          g_signal_emit_by_name (grabber, "shortcut-activated",
+                                 context.result, timestamp);
+        }
+    }
 
   gdk_display_flush (display);
   gdk_x11_display_error_trap_pop_ignored (display);
